import json
from copy import deepcopy
from pathlib import Path

import pytest

SPECS_ROOT_DIR = Path(__file__).absolute().parent / "data"
COMPANY_BASIC_INFO: dict = json.loads(
    (SPECS_ROOT_DIR / "CompanyBasicInfo.json").read_text(encoding="utf8")
)


@pytest.fixture
def company_basic_info():
    """
    Fixture that returns a deepcopy of the Company Basic Info to use in tests.
    The tests can freely modify it and corrupt it (easier than having multiple corrupt
    definitions in the repo).

    For performance reasons we load it once from filesystem (outside the fixture).
    """
    return deepcopy(COMPANY_BASIC_INFO)



import re
from typing import Any, Dict

from semver import Version

_MAJOR_MINOR_NO_PATCH_REGEX_TEMPLATE = r"""
        ^
        (?P<major>0|[1-9]\d*)
        (?:
            \.
            (?P<minor>0|[1-9]\d*)
        )
        (?:-(?P<prerelease>
            (?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)
            (?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*
        ))?
        (?:\+(?P<build>
            [0-9a-zA-Z-]+
            (?:\.[0-9a-zA-Z-]+)*
        ))?
        $
    """
_MAJOR_MINOR_NO_PATCH_REGEX = re.compile(
    _MAJOR_MINOR_NO_PATCH_REGEX_TEMPLATE.format(opt_patch=""),
    re.VERBOSE,
)


def parse_version_without_patch(version: str) -> Version:
    """
    Parse a version string that has no PATCH to a Version.

    :param version: version string
    :return: A Version
    :raise ValueError: If version is invalid

    >>> parse_version_without_patch('1.0')
    Version(major=1, minor=0, patch=0, prerelease=None, build=None)
    >>> parse_version_without_patch('1.0-beta.2')
    Version(major=1, minor=0, patch=0, prerelease='beta.2', build=None)
    """

    match = _MAJOR_MINOR_NO_PATCH_REGEX.match(version)
    if match is None:
        raise ValueError(f"{version} is not valid version string without patch")

    matched_version_parts: Dict[str, Any] = match.groupdict()
    matched_version_parts["patch"] = 0

    return Version(**matched_version_parts)



from typing import List

from pydantic import Field

from definition_tooling.converter import CamelCaseModel, DataProductDefinition


class CurrentAirQualityRequest(CamelCaseModel):
    lat: float = Field(
        ...,
        title="Latitude",
        description="The latitude coordinate of the desired location",
        examples=[60.192059],
        ge=-90,
        le=90,
    )
    lon: float = Field(
        ...,
        title="Longitude",
        description="The longitude coordinate of the desired location",
        examples=[24.945831],
        ge=-180,
        le=180,
    )


class CurrentAirQualityResponse(CamelCaseModel):
    air_quality_index: int = Field(
        ...,
        title="Air Quality Index",
        description=(
            "Current air quality index.\nRanges:\n0-50 Good;\n51-100 Moderate;\n"
            "101-150 Unhealthy For Sensitive Groups;\n151-200 Unhealthy;\n"
            "201-300 Very Unhealthy;\n301+ Hazardous"
        ),
        ge=0,
        examples=[30],
    )
    timestamp: str = Field(
        ...,
        title="Timestamp",
        description="Current timestamp in RFC 3339 format",
        examples=["2020-04-03T13:00:00Z"],
    )
    attribution: List[str] = Field(
        ...,
        title="Source Attribution",
        description="List of text to show required credits to data sources",
        examples=[["Moscow State environmental monitoring"]],
    )


DEFINITION = DataProductDefinition(
    version="1.0.0",
    title="Air Quality Index",
    description="Data Product for current air quality index",
    request=CurrentAirQualityRequest,
    response=CurrentAirQualityResponse,
)



class ValidatorError(Exception):
    def __init__(self, msg=None):
        if msg:
            self.msg = msg

    def __str__(self):
        return self.msg or ""


class OpenApiValidationError(ValidatorError):
    pass


class InvalidJSON(OpenApiValidationError):
    msg = "The file does not contain valid JSON"


class UnsupportedVersion(OpenApiValidationError):
    msg = "Validator supports only OpenAPI 3.x specs"


class MissingParameter(OpenApiValidationError):
    pass


class DefinitionError(OpenApiValidationError):
    pass


class WrongContentType(DefinitionError):
    msg = "Model description must be in application/json format"


class SchemaMissing(DefinitionError):
    pass


class MandatoryField(DefinitionError):
    pass


class NoEndpointsDefined(DefinitionError):
    pass


class OnlyOneEndpointAllowed(DefinitionError):
    pass


class PostMethodIsMissing(DefinitionError):
    pass


class OnlyPostMethodAllowed(DefinitionError):
    pass


class RequestBodyMissing(DefinitionError):
    pass


class ResponseBodyMissing(DefinitionError):
    pass


class AuthorizationHeaderMissing(DefinitionError):
    msg = "Authorization header is missing"


class AuthProviderHeaderMissing(DefinitionError):
    msg = "X-Authorization-Provider header is missing"


class ServersShouldNotBeDefined(DefinitionError):
    msg = '"servers" section should not exist in definition'


class SecurityShouldNotBeDefined(DefinitionError):
    msg = '"security" section should not exist in definition'


class HTTPResponseIsMissing(DefinitionError):
    pass


class VersionError(DefinitionError):
    pass


class InvalidOrMissingVersion(VersionError):
    msg = '"version" in "info" is missing or not a valid semantic version'


class UnexpectedVersionInFilename(VersionError):
    msg = "Test or draft definition should not have version in filename"


class TooHighVersion(VersionError):
    msg = "Test and draft definition should have version < 0.1.0"


class TooLowVersion(VersionError):
    msg = "Definition (not test or draft) should have version >= 0.1.0"


class InvalidOrMissingVersionInFileName(VersionError):
    msg = "The version number in the filename is missing or has incorrect format"


class VersionMissmatch(VersionError):
    msg = "The versions in the definition and filename do not match"



import json
from pathlib import Path

import pytest
from pydantic import ValidationError

from definition_tooling.converter import (
    CamelCaseModel,
    DataProductDefinition,
    convert_data_product_definitions,
)


def test_air_quality(tmpdir, json_snapshot):
    out_dir = tmpdir.mkdir("output")
    convert_data_product_definitions(Path(__file__).parent / "data", Path(out_dir))

    dest_file = out_dir / "AirQuality" / "Current.json"
    assert dest_file.exists()

    dest_spec = json.loads(dest_file.read_text("utf-8"))
    assert json_snapshot == dest_spec


def test_company_basic_info_errors(tmpdir, json_snapshot):
    """
    Test with a definition that includes custom error message
    """
    out_dir = tmpdir.mkdir("output")
    convert_data_product_definitions(Path(__file__).parent / "data", Path(out_dir))

    dest_file = out_dir / "Company" / "BasicInfo.json"
    assert dest_file.exists()

    dest_spec = json.loads(dest_file.read_text("utf-8"))
    assert json_snapshot == dest_spec


def test_current_weather_required_headers(tmpdir, json_snapshot):
    out_dir = tmpdir.mkdir("output")
    convert_data_product_definitions(Path(__file__).parent / "data", Path(out_dir))

    dest_file = out_dir / "Weather" / "Current" / "Metric.json"
    assert dest_file.exists()

    dest_spec = json.loads(dest_file.read_text("utf-8"))
    assert json_snapshot == dest_spec


def test_teapot_deprecated(tmpdir, json_snapshot):
    out_dir = tmpdir.mkdir("output")
    convert_data_product_definitions(Path(__file__).parent / "data", Path(out_dir))

    dest_file = out_dir / "Appliance" / "CoffeeBrewer.json"
    assert dest_file.exists()

    dest_spec = json.loads(dest_file.read_text("utf-8"))
    assert json_snapshot == dest_spec


def test_required_fields():
    with pytest.raises(ValidationError):
        DataProductDefinition(
            title=None,
            description="Description",
            request=CamelCaseModel,
            response=CamelCaseModel,
        )
    with pytest.raises(ValidationError):
        DataProductDefinition(
            title="Title",
            description=None,
            request=CamelCaseModel,
            response=CamelCaseModel,
        )


def test_summary_and_route_description(tmpdir, json_snapshot):
    out_dir = tmpdir.mkdir("output")
    convert_data_product_definitions(Path(__file__).parent / "data", Path(out_dir))

    dest_file = out_dir / "AirQuality" / "Current.json"
    assert dest_file.exists()

    dest_spec = json.loads(dest_file.read_text("utf-8"))

    title = dest_spec["info"]["title"]
    summary = dest_spec["paths"]["/AirQuality/Current"]["post"]["summary"]
    assert title == summary

    desc = dest_spec["info"]["description"]
    route_desc = dest_spec["paths"]["/AirQuality/Current"]["post"]["description"]
    assert desc == route_desc


def test_tags(tmpdir, json_snapshot):
    out_dir = tmpdir.mkdir("output")
    convert_data_product_definitions(Path(__file__).parent / "data", Path(out_dir))

    weather_file = out_dir / "Weather" / "Current" / "Metric.json"
    assert weather_file.exists()
    weather_spec = json.loads(weather_file.read_text("utf-8"))
    weather_tags = weather_spec["paths"]["/Weather/Current/Metric"]["post"]["tags"]
    # Check tags are alphabetically ordered and no duplicates
    assert weather_tags == ["humidity", "rain", "temperature", "wind"]

    coffee_file = out_dir / "Appliance" / "CoffeeBrewer.json"
    assert coffee_file.exists()
    coffee_spec = json.loads(coffee_file.read_text("utf-8"))
    coffee_tags = coffee_spec["paths"]["/Appliance/CoffeeBrewer"]["post"]["tags"]
    # Check tags are alphabetically ordered and no duplicates
    assert coffee_tags == ["brewer", "coffee"]



import json

import pytest

from definition_tooling.validator import errors as err
from definition_tooling.validator.core import DefinitionValidator


@pytest.fixture()
def spec(company_basic_info):
    """
    Make spec an alias of company_basic_info to shorten down the tests.
    """
    yield company_basic_info


@pytest.fixture()
def validate_definition(tmp_path):
    def validate_definition_(file: str, spec: dict):
        spec_path = tmp_path / f"{file}.json"
        spec_path.parent.mkdir(parents=True)
        spec_path.write_text(json.dumps(spec))
        DefinitionValidator(spec_path=spec_path, root_path=tmp_path).validate()

    return validate_definition_


@pytest.fixture()
def check_version_error(tmp_path):
    def check_version_error_(file: str, spec: dict, exception):
        spec_path = tmp_path / f"{file}.json"
        spec_path.parent.mkdir(parents=True)
        spec_path.write_text(json.dumps(spec))
        with pytest.raises(exception):
            DefinitionValidator(spec_path=spec_path, root_path=tmp_path).validate()

    return check_version_error_


def test_missing_version(check_version_error, spec):
    del spec["info"]["version"]
    check_version_error("Company/BasicInfo_v1.0", spec, err.InvalidOrMissingVersion)


@pytest.mark.parametrize(
    "version",
    [
        "1",
        "1.0",
        "1.0.0.0",
        "-1.0.0",
        "abc",
        None,
    ],
)
def test_invalid_version(check_version_error, spec, version: str):
    spec["info"]["version"] = version
    check_version_error("Company/BasicInfo_v1.0", spec, err.InvalidOrMissingVersion)


@pytest.mark.parametrize(
    "folder",
    [
        "draft",
        "test",
    ],
)
def test_version_in_filename_when_not_expected(check_version_error, spec, folder: str):
    spec["info"]["version"] = "0.0.1"
    check_version_error(
        f"{folder}/Company/BasicInfo_v0.0", spec, err.UnexpectedVersionInFilename
    )


@pytest.mark.parametrize(
    "folder,version",
    [
        ["draft", "1.0.0"],
        ["draft", "0.1.0"],
        ["test", "1.0.0"],
        ["test", "0.1.0"],
    ],
)
def test_too_high_version(check_version_error, spec, folder: str, version: str):
    spec["info"]["version"] = version
    check_version_error(f"{folder}/Company/BasicInfo", spec, err.TooHighVersion)


@pytest.mark.parametrize(
    "version,file_version",
    [
        ["0.0.1", "0.0"],
        ["0.0.2", "0.0"],
    ],
)
def test_too_low_version(check_version_error, spec, version: str, file_version: str):
    spec["info"]["version"] = version
    check_version_error(f"Company/BasicInfo_v{file_version}", spec, err.TooLowVersion)


@pytest.mark.parametrize(
    "filename",
    [
        "Company/BasicInfo",
        "Company/BasicInfo_vabc",
        "Company/BasicInfo_v1",
        "Company/BasicInfo_v1.1.0",
        "Company/BasicInfo_v_1.0",
        "Company/BasicInfo_1.1",
    ],
)
def test_incorrect_filename_version(check_version_error, spec, filename: str):
    check_version_error(filename, spec, err.InvalidOrMissingVersionInFileName)


@pytest.mark.parametrize(
    "version,file_version",
    [
        ["1.0.0", "1.1"],
        ["1.1.0", "1.0"],
        ["2.0.0", "1.0"],
    ],
)
def test_version_missmatch(check_version_error, spec, version: str, file_version: str):
    spec["info"]["version"] = version
    check_version_error(
        f"Company/BasicInfo_v{file_version}", spec, err.VersionMissmatch
    )


@pytest.mark.parametrize(
    "version,filename",
    [
        ["0.0.1", "test/Company/BasicInfo"],
        ["0.0.2", "test/Company/BasicInfo"],
        ["0.0.1", "draft/Company/BasicInfo"],
        ["0.0.200", "draft/Company/BasicInfo"],
        ["0.1.0", "Company/BasicInfo_v0.1"],
        ["0.1.1", "Company/BasicInfo_v0.1"],
        ["0.1.2", "Company/BasicInfo_v0.1"],
        ["1.0.0", "Company/BasicInfo_v1.0"],
        ["1.1.0-beta", "Company/BasicInfo_v1.1-beta"],
        ["1.1.0-beta.2", "Company/BasicInfo_v1.1-beta.2"],
        ["1.1.0", "Company/BasicInfo_v1.1"],
        ["12.34.56", "Company/BasicInfo_v12.34"],
    ],
)
def test_valid_versions(validate_definition, spec, version: str, filename: str):
    spec["info"]["version"] = version
    validate_definition(filename, spec)



"""
Predefined errors that the product gateway or productizers can return.

These errors can not be overridden by the data product definition itself.
"""

from pydantic import BaseModel, Field


class BaseApiError(BaseModel):
    __status__: int

    @classmethod
    def get_response_spec(cls):
        return {"model": cls}


class ApiError(BaseApiError):
    type: str = Field(..., title="Error type", description="Error identifier")
    message: str = Field(..., title="Error message", description="Error description")


class ApiOrExternalError(BaseApiError):
    @classmethod
    def get_response_spec(cls):
        return {"model": cls, "content": {"text/plain": {}, "text/html": {}}}


class Unauthorized(ApiError):
    __status__ = 401


class Forbidden(ApiError):
    __status__ = 403


class NotFound(ApiError):
    __status__ = 404


# Note: 422 is added automatically by FastAPI


class RateLimitExceeded(BaseApiError):
    """
    This response is reserved by Product Gateway.
    """

    __status__ = 429
    message: str = Field(
        "Rate limit exceeded",
        title="Error message",
        description="Error description",
    )


class DataSourceNotFound(BaseApiError):
    """
    This response is reserved by Product Gateway.
    """

    __status__ = 444
    message: str = Field(
        "Data source not found",
        title="Error message",
        description="Error description",
    )


class DataSourceError(ApiError):
    __status__ = 500


class BadGateway(BaseApiError):
    """
    This response is reserved by Product Gateway.
    """

    __status__ = 502


class ServiceUnavailable(ApiOrExternalError):
    """
    This response is reserved by Product Gateway.
    """

    __status__ = 503
    message: str = Field("", title="Error message", description="Error description")


class GatewayTimeout(ApiOrExternalError):
    """
    This response is reserved by Product Gateway.
    """

    __status__ = 504
    message: str = Field(
        "",
        title="Error message",
        description="Error description",
    )


class DoesNotConformToDefinition(BaseApiError):
    """
    This response is reserved by Product Gateway.
    """

    __status__ = 550
    message: str = "Response from data source does not conform to definition"
    data_source_status_code: int = Field(
        ...,
        title="Data source status code",
        description="HTTP status code returned from the data source",
    )


DATA_PRODUCT_ERRORS = {
    resp.__status__: resp.get_response_spec()
    for resp in [
        Unauthorized,
        Forbidden,
        NotFound,
        RateLimitExceeded,
        DataSourceNotFound,
        DataSourceError,
        BadGateway,
        ServiceUnavailable,
        GatewayTimeout,
        DoesNotConformToDefinition,
    ]
}



import json
from pathlib import Path
from typing import Union

from semver import Version

from definition_tooling.api_errors import DATA_PRODUCT_ERRORS
from definition_tooling.validator import errors as err
from definition_tooling.validator.utils import parse_version_without_patch


def validate_component_schema(spec: dict, components_schema: dict):
    if not spec["content"].get("application/json"):
        raise err.WrongContentType
    ref = spec["content"]["application/json"].get("schema", {}).get("$ref")
    if not ref:
        raise err.SchemaMissing(
            'Request or response model is missing from "schema/$ref" section'
        )
    if not ref.startswith("#/components/schemas/"):
        raise err.SchemaMissing(
            "Request and response models must be defined in the"
            '"#/components/schemas/" section'
        )
    model_name = ref.split("/")[-1]
    if not components_schema.get(model_name):
        raise err.SchemaMissing(f"Component schema is missing for {model_name}")


def validate_version(spec: dict, spec_path: Path, root_path: Path):
    """
    Validate version in definition and filename.

    Definitions in "test/" and "draft/" namespaces should:
    - have versions < 0.1.0
    - not have any version in the filename

    All other definitions should:
    - have version >= 0.1.0
    - have the _v{MAJOR}.{MINOR} number in the filename
    - have versions that match the version in the filename

    :param spec: OpenAPI spec
    :param spec_path: Path to the actual definition
    :param root_path: Path to the root of definitions
    :raises VersionError: When there's a problem with the version number.
    """
    try:
        version = Version.parse(spec["info"]["version"])
    except (ValueError, TypeError, KeyError):
        raise err.InvalidOrMissingVersion

    test_definition = spec_path.is_relative_to(root_path / "test")
    draft_definition = spec_path.is_relative_to(root_path / "draft")
    _, __, file_version_str = spec_path.stem.partition("_v")

    if test_definition or draft_definition:
        if file_version_str:
            raise err.UnexpectedVersionInFilename
        if version >= Version.parse("0.1.0"):
            raise err.TooHighVersion
    else:
        try:
            file_version = parse_version_without_patch(file_version_str)
        except ValueError:
            raise err.InvalidOrMissingVersionInFileName

        if file_version != Version(
            version.major, version.minor, 0, version.prerelease, version.build
        ):
            raise err.VersionMissmatch

        if version < Version.parse("0.1.0"):
            raise err.TooLowVersion


def validate_spec(spec: dict, spec_path: Path, root_path: Path):
    """
    Validate that OpenAPI spec looks like a data product definition. For example, that
    it only has one POST method defined.

    :param spec: OpenAPI spec
    :param spec_path: Path to the actual definition
    :param root_path: Path to the root of specs
    :raises OpenApiValidationError: When OpenAPI spec is incorrect
    """
    if "servers" in spec:
        raise err.ServersShouldNotBeDefined

    if not spec.get("openapi", "").startswith("3."):
        raise err.UnsupportedVersion("Validator supports only OpenAPI 3.x specs")

    for field in {"title", "description"}:
        if not spec.get("info", {}).get(field):
            raise err.MandatoryField(f'{field} is a required field in "info" section')

    paths = spec.get("paths", {})
    if not paths:
        raise err.NoEndpointsDefined
    if len(paths) > 1:
        raise err.OnlyOneEndpointAllowed

    post_route = {}
    for name, path in paths.items():
        methods = list(path)
        if "post" not in methods:
            raise err.PostMethodIsMissing
        if methods != ["post"]:
            raise err.OnlyPostMethodAllowed
        post_route = path["post"]

    for field in {"summary", "description"}:
        if not post_route.get(field):
            raise err.MandatoryField(f"{field} is a required field for POST route")

    component_schemas = spec.get("components", {}).get("schemas")
    if not component_schemas:
        raise err.SchemaMissing('No "components/schemas" section defined')

    if "security" in post_route:
        raise err.SecurityShouldNotBeDefined

    if post_route.get("requestBody", {}).get("content"):
        validate_component_schema(post_route["requestBody"], component_schemas)

    responses = post_route.get("responses", {})
    if not responses.get("200") or not responses["200"].get("content"):
        raise err.ResponseBodyMissing
    validate_component_schema(responses["200"], component_schemas)

    for code in {*DATA_PRODUCT_ERRORS, 422}:
        if not responses.get(str(code)):
            raise err.HTTPResponseIsMissing(f"Missing response for status code {code}")

    headers = [
        param.get("name", "").lower()
        for param in post_route.get("parameters", [])
        if param.get("in") == "header"
    ]
    if "authorization" not in headers:
        raise err.AuthorizationHeaderMissing
    if "x-authorization-provider" not in headers:
        raise err.AuthProviderHeaderMissing

    validate_version(spec=spec, spec_path=spec_path, root_path=root_path)


class DefinitionValidator:
    def __init__(
        self,
        spec_path: Union[str, Path],
        root_path: Union[str, Path],
    ):
        self.root_path = Path(root_path)
        self.spec_path = Path(spec_path)

    def validate(self):
        try:
            spec = json.loads(self.spec_path.read_text(encoding="utf8"))
        except json.JSONDecodeError:
            raise err.InvalidJSON(f"Incorrect JSON: {self.spec_path}")
        except Exception as e:
            raise err.ValidatorError(f"Failed to validate {self.spec_path}: {e}")
        self.validate_spec(spec, spec_path=self.spec_path, root_path=self.root_path)

    @classmethod
    def validate_spec(cls, spec: dict, spec_path: Path, root_path: Path):
        # it's moved to separate function to reduce indentation
        return validate_spec(spec, spec_path=spec_path, root_path=root_path)
